from app.application import list_tables
from app.database import delete_table
from app.database import initialize
from app.database import insert_into_table
from app.database import rename_table
from app.utility import convert_time_to_epoch


def upgrade_0_5_1(cur):

    for table in list_tables():
        temp_table = "temp_" + table
        initialize(temp_table)

        query = f"SELECT * FROM {table}"
        res = cur.execute(query).fetchall()

        for each in res:
            deadline = each[5]
            completed = each[8]
            parent_id = each[2]
            if deadline == "None":
                deadline = "0"
            else:
                deadline = str(convert_time_to_epoch(deadline))

            if completed == "None":
                completed = "0"
            else:
                completed = str(convert_time_to_epoch(completed))

            if parent_id == None:
                parent_id = "NULL"

            task = (
                str(each[0]),
                f"'{each[1]}'",
                str(parent_id),
                f"'{each[3]}'",
                f"'{each[4]}'",
                deadline,
                str(each[6]),
                f"'{each[7]}'",
                completed,
                str(each[9]),
            )
            columns = [
                "id",
                "title",
                "parent_id",
                "description",
                "status",
                "deadline",
                "priority",
                "label",
                "completed",
                "subtasks",
            ]
            insert_into_table(temp_table, columns, task)

        delete_table(table)
        rename_table(temp_table, table)


def upgrade_0_6_3(cur):
    """
    Upgrade the database to version 0.6.3
    """
    for table in list_tables():
        # Drop the existing trigger
        cur.execute(f"DROP TRIGGER IF EXISTS initialize_completed_column_{table};")

        # Add a new trigger to check the value of completed before updating it
        cur.execute(
            f"""
            CREATE TRIGGER initialize_completed_column_{table}
            AFTER INSERT ON {table}
            FOR EACH ROW
            BEGIN
                UPDATE {table} SET completed = NEW.deadline WHERE id = NEW.id AND NEW.completed == 0;
            END;
            """,
        )


def upgrade_1_0_0(cur):
    """
    Upgrade the database to version 1.0.0
    """
    cur.execute(
        """
        CREATE TABLE IF NOT EXISTS sessions (
            session_id INTEGER PRIMARY KEY AUTOINCREMENT,
            task_id INTEGER,
            table_name VARCHAR NOT NULL,
            start_datetime INTEGER NOT NULL,
            end_datetime INTEGER NOT NULL
        )""",
    )
    cur.execute(
        """
        CREATE TABLE IF NOT EXISTS session_data (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            session_id INTEGER NOT NULL,
            application_name VARCHAR NOT NULL,
            duration INTEGER NOT NULL,
            FOREIGN KEY(session_id) REFERENCES sessions(session_id)
            ON DELETE CASCADE
        )""",
    )
    cur.execute(
        f"CREATE INDEX IF NOT EXISTS session_id_session on sessions(session_id);",
    )
    cur.execute(
        f"CREATE INDEX IF NOT EXISTS session_id_data on session_data(session_id);",
    )
    tables = list_tables()
    for table in tables:
        # Drop the existing trigger
        cur.execute(f"DROP INDEX IF EXISTS title_{table};")

        # Add a new trigger to check the value of completed before updating it
        cur.execute(
            f"""
            CREATE INDEX IF NOT EXISTS index_{table} ON {table}(id, title);
            """,
        )



import platform
import time
from datetime import datetime
from datetime import timedelta

from click import echo
from click import style


def convert_time_to_epoch(time_str, eod=True):
    if time_str in ("None", None):
        return 0
    separator_list = [":", ".", "-", "/", "\\", "|", "_", ","]
    for separator in separator_list:
        if separator in time_str:
            date_list = time_str.split(separator)
            break
    else:
        return "No separator found"
    if len(date_list) != 3:
        return "Length not 3"

    if len(date_list[0]) != 4:
        date_list = date_list[::-1]

    if eod == True:
        hr = 23
        minute = 59
        sec = 59
    else:
        hr = 0
        minute = 0
        sec = 0

    try:
        date_obj = datetime(
            int(date_list[0]),
            int(date_list[1]),
            int(date_list[2]),
            hour=hr,
            minute=minute,
            second=sec,
        )
        epoch_time = int(time.mktime(date_obj.timetuple()))
        return epoch_time
    except Exception as e:
        return e.__str__()


def convert_epoch_to_date(epoch_time):
    if epoch_time == 0:
        return "None"
    try:
        date_obj = datetime.fromtimestamp(epoch_time)
        return date_obj.strftime("%d-%m-%Y")
    except Exception as e:
        return e.__str__()


def convert_epoch_to_datetime(epoch_time):
    if epoch_time == 0:
        return "None"
    try:
        date_obj = datetime.fromtimestamp(epoch_time)
        return date_obj.strftime("%H:%M, %-d/%-m/%Y")
    except Exception:
        return "invalid"


def convert_seconds_delta_to_time(seconds):
    seconds = int(seconds)
    hrs = seconds // 3600
    mins = (seconds % 3600) // 60
    secs = seconds % 60
    parts = []
    if hrs > 0:
        parts.append(f"{hrs} hrs")
    if mins > 0:
        parts.append(f"{mins} mins")
    if secs > 0 or not parts:
        parts.append(f"{secs} secs")
    return ", ".join(parts)


def sanitize_text(text):
    return text.strip().replace("'", '"')


def generate_migration_error():
    display_error_message(
        "Have You Run Migrations? Run 'devcord init --migrate' to run migrations",
    )


def display_error_message(message: str):
    """
    Display an error message in red style with an error prefix.
    """
    echo(
        style(
            f"Error: {message}",
            fg="red",
        ),
    )


def display_info_message(message: str):
    """
    Display an informational message in green style with an info prefix.
    """
    echo(
        style(
            f"Info: {message}",
            fg="yellow",
        ),
    )


def display_success_message(message: str):
    """
    Display a success message in green style with a success prefix.
    """
    echo(
        style(
            f"Success: {message}",
            fg="green",
        ),
    )


def sanitize_table_name(table_name: str) -> (str, bool):  # type: ignore
    for ch in table_name:
        if ch.isalnum() or ch == " " or ch == "_":
            continue
        else:
            return "", False

    return table_name.replace(" ", "_"), True


def get_weekend() -> str:
    """
    Get the date of the current week's end (Sunday).
    """
    today = datetime.now()
    days_until_sunday = 6 - today.weekday()
    sunday = today + timedelta(days=days_until_sunday)
    return sunday.strftime("%d-%m-%Y")


def get_week_start() -> str:
    """
    Get the date of the current week's start (Monday).
    """
    today = datetime.now()
    days_since_monday = today.weekday()
    monday = today - timedelta(days=days_since_monday)
    return monday.strftime("%d-%m-%Y")


def get_relative_date_string(relative_days: int) -> str:
    """
    Get the date string for today + relative_days.

    Args:
        relative_days (int): Number of days relative to today. Can be positive or negative.

    Returns:
        str: Date string in the format "dd-mm-yyyy".
    """
    target_date = datetime.now() + timedelta(days=relative_days)
    return target_date.strftime("%d-%m-%Y")


def check_if_relative_deadline(deadline: str) -> str:
    if deadline[0] == "+" or deadline[0] == "-":
        if deadline[1:].isdigit():
            deadline = get_relative_date_string(int(deadline))
        else:
            display_error_message("Example: '+4'")
            return False
    return deadline


def get_os():
    system = platform.system()
    if system == "Linux":
        return "Linux"
    elif system == "Darwin":
        return "MacOS"
    elif system == "Windows":
        return "Windows"
    else:
        return "Unknown"



import curses

from click import echo
from click import style

from app import application
from app.console_helper import treeify
from app.constants import CR_ENTER
from app.constants import LF_ENTER
from app.utility import display_error_message
from app.utility import sanitize_table_name


def check_table_exists(table_name: str) -> bool:
    table_name, ok = sanitize_table_name(table_name)
    if not ok:
        display_error_message(
            "Table name is not valid, please use only alphanumeric characters or underscores."
            + "Maybe you are not a developer?",
        )
    return table_name in application.list_tables(), table_name


def lister(table, completed=False):
    tasks = application.list_tasks(table, subtasks=True, completed=completed)

    if len(tasks) == 0:
        return None

    tree = treeify(tasks)

    def get_tasks(task_id):
        if task_id == None:
            return {}, []
        if task_id == -1:
            return {"title": table, "id": -1}, tree[-1]["children"]
        return tree[task_id]["data"], tree[task_id]["children"]

    return curses.wrapper(menu, -1, get_tasks)


def session_lister(table, task_id=None):
    sessions = application.list_sessions(table, task_id=task_id)
    if len(sessions) == 0:
        return None

    def get_sessions(id: int = -1):
        if id == -1:
            return_list = []
            for session in sessions:
                task_name = session["task_name"]
                if len(session["task_name"]) > 20:
                    task_name = session["task_name"][:20] + "..."
                return_list.append(
                    {
                        "data": {
                            "id": session["session_id"],
                            "title": f"{task_name} {session['start_datetime']} -> {session['end_datetime']}",
                        },
                        "children": [],
                    },
                )
            return {"title": "Sessions", "id": -1}, return_list
        else:
            return {}, []

    return curses.wrapper(menu, -1, get_sessions)


def menu(stdscr, current: int, get_tasks) -> dict:
    """
    'current' is the id of the current task;
    'get_tasks' is a function that takes an id and returns the task and its subtasks.
    If id is None, it returns None, and empty list.
    If id is -1, it returns root task and its subtasks.
    """
    curses.curs_set(0)
    stdscr.keypad(1)

    selected = 0
    current_task, _ = get_tasks(current)  # get initial list of tasks
    parent, options_left = get_tasks(
        current_task.get("parent_id", None),
    )  # for one page of tasks, parent is the same so get them once before hand

    while True:
        stdscr.clear()
        current_task, options = get_tasks(current)
        title = current_task["title"]

        height, width = stdscr.getmaxyx()
        # Leave space for top/bottom margins
        height -= 2
        width -= 1

        if len(title) > 30:
            title = title[:30] + "..."

        start_y = 1
        start_x = 1
        stdscr.addstr(start_y, start_x, f"{title}:")
        if len(options_left) > 0:
            stdscr.addstr(start_y + 1, start_x + 1, "<<-- prev")

        if len(options) > 0:
            _, options_right = get_tasks(options[selected]["data"]["id"])
            if len(options_right) > 0:
                stdscr.addstr(start_y + 1, start_x + 15, "next -->>")

        delta_x = 2
        delta_y = start_y + 3

        max_page_size = height - delta_y + 1

        idx = 0
        page_size = min(max_page_size, len(options))
        if len(options) > max_page_size:
            idx = selected
            page_size = min(max_page_size, len(options) - selected) + selected
            delta_y -= selected

        while idx < page_size:
            option = options[idx]
            title = option["data"]["title"]
            if len(title) > width - 2:
                title = title[: width - 5] + "..."

            if idx == selected:
                stdscr.addstr(
                    idx + delta_y,
                    delta_x,
                    title,
                    curses.A_REVERSE,
                )
            else:
                stdscr.addstr(idx + delta_y, delta_x, title)
            idx += 1

        key = stdscr.getch()

        if key == curses.KEY_UP:
            selected = (selected - 1) % len(options)

        elif key == curses.KEY_DOWN:
            selected = (selected + 1) % len(options)

        elif key == curses.KEY_RIGHT and len(options_right) > 0:
            current = options[selected]["data"]["id"]
            options_left = options
            parent = current_task
            options = options_right
            selected = 0

        elif key == curses.KEY_LEFT and len(options_left) > 0:
            options = options_left
            for idx, option in enumerate(options):
                if option["data"]["id"] == current:
                    selected = idx
                    break
            current = parent["id"]
            parent, options_left = get_tasks(current_task.get("parent_id", None))

        elif key in [curses.KEY_ENTER, LF_ENTER, CR_ENTER] and len(options) > 0:
            return options[selected]["data"]



import sqlite3

from app.constants import db_path

# we should have separate functions to create triggers and indexes


def initialize(table_name: str) -> None:
    """
    Initialize the database with all the necessary tables.
    """
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    cur.execute(
        f"""CREATE TABLE {table_name}(
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        title VARCHAR NOT NULL, parent_id INTEGER,
        description TEXT DEFAULT 'None',
        status VARCHAR DEFAULT 'Pending',
        deadline INTEGER DEFAULT 0,
        priority INTEGER DEFAULT 0,
        label VARCHAR DEFAULT 'None',
        completed INTEGER DEFAULT 0,
        subtasks INTEGER DEFAULT 0
        )""",
    )
    cur.execute(
        f"""
        CREATE TRIGGER initialize_completed_column_{table_name}
        AFTER INSERT ON {table_name}
        FOR EACH ROW
        BEGIN
            UPDATE {table_name} SET completed = NEW.deadline WHERE id = NEW.id AND NEW.completed == 0;
        END;
    """,
    )
    cur.execute(
        f"CREATE INDEX IF NOT EXISTS index_{table_name} ON {table_name}(id, title);",
    )
    cur.execute(
        f"""
        CREATE TABLE IF NOT EXISTS sessions (
            session_id INTEGER PRIMARY KEY AUTOINCREMENT,
            task_id INTEGER,
            table_name VARCHAR NOT NULL,
            start_datetime INTEGER NOT NULL,
            end_datetime INTEGER NOT NULL
        )""",
    )
    cur.execute(
        f"""
        CREATE TABLE IF NOT EXISTS session_data (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            session_id INTEGER NOT NULL,
            application_name VARCHAR NOT NULL,
            duration INTEGER NOT NULL,
            FOREIGN KEY(session_id) REFERENCES sessions(session_id)
            ON DELETE CASCADE
        )""",
    )
    cur.execute(
        f"CREATE INDEX IF NOT EXISTS session_id_session on sessions(session_id);",
    )
    cur.execute(
        f"CREATE INDEX IF NOT EXISTS session_id_data on session_data(session_id);",
    )
    conn.commit()


def list_tables() -> list:
    """
    List all the tables in the database.
    """
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    res = cur.execute("SELECT name FROM sqlite_master WHERE type='table';").fetchall()
    conn.commit()
    return res


def list_table(
    table: str,
    columns: list,
    where_clause: str = "",
    group_clause: str = "",
    order_by: str = "",
) -> list:
    """
    List data from a table.
    """
    query = f"SELECT {', '.join(columns)} FROM {table} {where_clause} {group_clause} {order_by}"
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    res = cur.execute(query).fetchall()
    conn.commit()
    return res


def insert_into_table(table: str, columns: list, values: list) -> None:
    """
    Insert data into a table.
    """
    query = f"INSERT INTO {table} ({', '.join(columns)}) VALUES ({', '.join(values)})"
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    cur.execute(query)
    conn.commit()


def update_table(table: str, new_data: dict) -> None:
    """
    Update Values Of Given Table
    """
    set_clause = ", ".join(
        [f"{key} = {value}" for key, value in new_data.items() if key != "id"],
    )
    query = f"UPDATE {table} SET {set_clause} WHERE id = {new_data['id']}"
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    cur.execute(query)
    conn.commit()


def delete_task(table: str, task_id: int) -> None:
    """
    Delete Task From Database
    """
    query = f"DELETE FROM {table} WHERE id = {task_id}"
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    cur.execute(query)
    conn.commit()


def delete_table(table_name: str) -> None:
    """
    Delete Table From Database
    """
    query = f"DROP TABLE {table_name}; "
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    cur.execute(query)
    cur.execute(f"DROP INDEX IF EXISTS title_{table_name}")
    cur.execute(f"DROP TRIGGER IF EXISTS initialize_completed_column_{table_name};")
    conn.commit()


def rename_table(old_table_name: str, new_table_name: str) -> None:
    """
    Rename Table In Database
    """
    query = f"ALTER TABLE {old_table_name} RENAME TO {new_table_name}"
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    cur.execute(query)
    try:
        cur.execute(
            f"ALTER INDEX title_{old_table_name} RENAME TO title_{new_table_name};",
        )
        cur.execute(
            f"ALTER TRIGGER initialize_completed_column_{old_table_name} RENAME TO initialize_completed_column_{new_table_name};",
        )
    except:
        # the code will reach here probably because of the index, so for now as a temp fix we will create the index instead of throwing an error.
        # will remove this in the later versions
        cur.execute(f"CREATE INDEX title_{new_table_name} on {new_table_name}(title);")
    conn.commit()


def add_session(
    task_id: int,
    table_name: str,
    start_datetime: int,
    end_datetime: int,
) -> int:
    """
    Add session to the database and return the row that was just created.
    """
    query = f"""
        INSERT INTO sessions(task_id, table_name, start_datetime, end_datetime)
        VALUES ({task_id}, '{table_name}', {start_datetime}, {end_datetime})
    """
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    cur.execute(query)
    session_id = cur.lastrowid  # Get the ID of the last inserted row
    conn.commit()
    return session_id


def add_session_data(
    session_id: int,
    application_name: str,
    duration: int,
) -> None:
    """
    Add session data to the database.
    """
    query = f"""
        INSERT INTO session_data (session_id, application_name, duration)
        VALUES ({session_id}, '{application_name}', {duration})
    """
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    cur.execute(query)
    conn.commit()


def list_sessions(table_name: str, task_id: int = None) -> list:
    """
    List sessions from the sessions table, optionally filtered by task_id, ordered by end_datetime.
    """
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    if task_id is not None:
        query = f"SELECT * FROM sessions WHERE table_name = ? AND task_id = ? ORDER BY end_datetime DESC"
        res = cur.execute(query, (table_name, task_id)).fetchall()
    else:
        query = (
            f"SELECT * FROM sessions WHERE table_name = ? ORDER BY end_datetime DESC"
        )
        res = cur.execute(query, (table_name,)).fetchall()
    conn.commit()
    return res


def get_session_data(session_id: int) -> list:
    """
    Get session data from the session_data table.
    """
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    query = f"SELECT * FROM session_data WHERE session_id = ?"
    res = cur.execute(query, (session_id,)).fetchall()
    conn.commit()
    return res


def delete_session(session_id: int) -> None:
    """
    Delete a session from the sessions table.
    """
    query = f"DELETE FROM sessions WHERE session_id = ?"
    conn = sqlite3.connect(db_path)
    cur = conn.cursor()
    cur.execute(query, (session_id,))
    conn.commit()



import json
import os
import signal
import sys
import tempfile
from datetime import datetime
from multiprocessing import Process

from ewmh import EWMH
from Xlib import display
from Xlib import X
from Xlib.protocol.event import PropertyNotify


def start_logging(file_handle: str):

    # Dictionary to store window name and active time
    activity = {}

    def trim_spaces_from_list(lst):
        return [item.strip() for item in lst if item.strip()]

    # Connect to the X server
    disp = display.Display()
    root = disp.screen().root
    ewmh = EWMH(disp)

    # Listen for property changes on the root window
    root.change_attributes(event_mask=X.PropertyChangeMask)

    # Key to get active window event
    NET_ACTIVE_WINDOW = disp.intern_atom("_NET_ACTIVE_WINDOW")

    last_window = ewmh.getActiveWindow()
    last_window_name = ewmh.getWmName(last_window)
    if isinstance(last_window_name, bytes):
        last_window_name = last_window_name.decode("utf-8")
    last_window_id = last_window.id
    last_switch_time = datetime.now()
    activity[last_window.id] = {
        "time": 0,
        "name_list": trim_spaces_from_list(list(last_window_name.split("-"))),
    }

    def add_time_to_current_window():
        current_time = datetime.now()
        name_list = trim_spaces_from_list(list(last_window_name.split("-")))
        if activity.get(last_window_id, None):
            activity[last_window_id]["time"] += (
                current_time - last_switch_time
            ).total_seconds()
            new_name_list = []
            for item in activity[last_window_id]["name_list"]:
                if item in name_list:
                    new_name_list.append(item)
            activity[last_window_id]["name_list"] = new_name_list
        else:
            activity[last_window_id] = {
                "time": (current_time - last_switch_time).total_seconds(),
                "name_list": name_list,
            }

    def handle_exit(signum, frame):
        try:
            add_time_to_current_window()
            with open(file_handle, "w+") as f:
                json.dump(activity, f, indent=4)
            print(f"Successfully captured session data.")
        except Exception as e:
            print(f"Failed to capturing session data: {e}")
        sys.exit(0)

    signal.signal(signal.SIGINT, handle_exit)
    signal.signal(signal.SIGTERM, handle_exit)

    while True:
        event = disp.next_event()
        if isinstance(event, PropertyNotify) and event.atom == NET_ACTIVE_WINDOW:

            active_window = ewmh.getActiveWindow()
            window_name = ewmh.getWmName(active_window)
            if isinstance(window_name, bytes):
                window_name = window_name.decode("utf-8")

            if active_window and active_window.id > 0 and window_name:
                add_time_to_current_window()
                if (datetime.now() - last_switch_time).total_seconds() > 30:
                    with open(file_handle, "w+") as f:
                        json.dump(activity, f, indent=4)

                last_window = active_window
                last_window_name = ewmh.getWmName(last_window)
                if isinstance(last_window_name, bytes):
                    last_window_name = last_window_name.decode("utf-8")
                last_window_id = last_window.id
                last_switch_time = datetime.now()


def session_start():
    # Create and start a process for start_logging
    tmp = tempfile.NamedTemporaryFile(delete=False)
    logging_process = Process(target=start_logging, args=(tmp.name,))
    logging_process.start()
    return logging_process.pid, tmp.name  # Return the process ID


def session_end(pid):
    try:
        os.kill(pid, signal.SIGTERM)
        return True, ""
    except ProcessLookupError:
        return False, None
    except Exception as e:
        return False, str(e)



import unittest
from datetime import datetime, timedelta
import time
from app.utility import convert_time_to_epoch, convert_epoch_to_date, get_week_start, get_weekend, get_relative_date_string, check_if_relative_deadline, convert_epoch_to_datetime, convert_seconds_delta_to_time

class ConvertToEpoch(unittest.TestCase):
    def test_positive(self):
        date_obj = datetime(2020, 12, 31, hour=23, minute=59, second=59)
        epoch_time = int(time.mktime(date_obj.timetuple()))
        self.assertEqual(convert_time_to_epoch("2020-12-31"), epoch_time)
        self.assertEqual(convert_time_to_epoch("2020/12/31"), epoch_time)
        self.assertEqual(convert_time_to_epoch("31.12.2020"), epoch_time)
        self.assertEqual(convert_time_to_epoch("31-12-2020"), epoch_time)
        self.assertEqual(convert_time_to_epoch("31/12/2020"), epoch_time)
        self.assertEqual(convert_time_to_epoch("31\\12\\2020"), epoch_time)
        self.assertEqual(convert_time_to_epoch("31|12|2020"), epoch_time)
        self.assertEqual(convert_time_to_epoch("31_12_2020"), epoch_time)
        self.assertEqual(convert_time_to_epoch("31,12,2020"), epoch_time)
        date_obj = datetime(2020, 12, 31, hour=0, minute=0, second=0)
        epoch_time = int(time.mktime(date_obj.timetuple()))
        self.assertEqual(convert_time_to_epoch("2020-12-31", False), epoch_time)
    
    def test_negative(self):
        self.assertEqual(convert_time_to_epoch("2020-12-31-12"), "Length not 3")
        self.assertEqual(convert_time_to_epoch("20020100"),"No separator found")
        self.assertEqual(convert_time_to_epoch("20-00-03"),"month must be in 1..12")
        self.assertEqual(convert_time_to_epoch("20-13-03"),"month must be in 1..12")
        self.assertEqual(convert_time_to_epoch("20-12-00"),"year 0 is out of range")
        self.assertEqual(convert_time_to_epoch("a-b-c"),"invalid literal for int() with base 10: 'c'")
        
class ConvertEpochToTime(unittest.TestCase):
    def test_positive(self):
        date_str = "31-12-2020"
        epoch = convert_time_to_epoch(date_str)
        self.assertEqual(convert_epoch_to_date(epoch), date_str)
    
    def test_negative(self):
        self.assertEqual(convert_epoch_to_date(0), "None")
        self.assertEqual(convert_epoch_to_date("abc"), "'str' object cannot be interpreted as an integer")
        
class GetWeekEnds(unittest.TestCase):
    def test_positive(self):
        self.assertEqual(get_weekend(), "01-09-2024")
        self.assertEqual(get_week_start(), "26-08-2024")
        
class TestGetRelativeDateString(unittest.TestCase):
    def test_positive_days(self):
        today = datetime.now()
        expected_date = (today + timedelta(days=5)).strftime("%d-%m-%Y")
        self.assertEqual(get_relative_date_string(5), expected_date)

    def test_negative_days(self):
        today = datetime.now()
        expected_date = (today - timedelta(days=3)).strftime("%d-%m-%Y")
        self.assertEqual(get_relative_date_string(-3), expected_date)

    def test_zero_days(self):
        today = datetime.now().strftime("%d-%m-%Y")
        self.assertEqual(get_relative_date_string(0), today)


class TestCheckIfRelativeDeadline(unittest.TestCase):
    def test_positive_relative_deadline(self):

        deadline = "+5"
        result = check_if_relative_deadline(deadline=deadline)
        expected_date = (datetime.now() + timedelta(days=5)).strftime("%d-%m-%Y")
        self.assertEqual(result, expected_date)

    def test_negative_relative_deadline(self):

        deadline = "-3"
        result = check_if_relative_deadline(deadline=deadline)
        expected_date = (datetime.now() - timedelta(days=3)).strftime("%d-%m-%Y")
        self.assertEqual(result, expected_date)

    def test_invalid_relative_deadline(self):

        deadline = "+abc"
        result = check_if_relative_deadline(deadline=deadline)
        self.assertFalse(result)

    def test_non_relative_deadline(self):

        deadline = "31-12-2025"
        result = check_if_relative_deadline(deadline=deadline)
        self.assertEqual(result, "31-12-2025")

class TestConvertEpochToDatetime(unittest.TestCase):
    def test_positive(self):
        date_obj = datetime(2020, 12, 31, hour=23, minute=59, second=59)
        epoch_time = int(time.mktime(date_obj.timetuple()))
        # The output format is "%H:%M, %-d/%-m/%Y"; handle platform differences for day/month padding
        result = convert_epoch_to_datetime(epoch_time)
        expected = date_obj.strftime("%H:%M, %-d/%-m/%Y")
        self.assertEqual(result, expected)

    def test_zero(self):
        self.assertEqual(convert_epoch_to_datetime(0), "None")

    def test_invalid(self):
        self.assertIn("invalid", convert_epoch_to_datetime("abc"))

class TestConvertSecondsDeltaToTime(unittest.TestCase):
    def test_hours_minutes_seconds(self):
        self.assertEqual(convert_seconds_delta_to_time(3661), "1 hrs, 1 mins, 1 secs")

    def test_only_minutes(self):
        self.assertEqual(convert_seconds_delta_to_time(120), "2 mins")

    def test_only_seconds(self):
        self.assertEqual(convert_seconds_delta_to_time(59), "59 secs")

    def test_zero(self):
        self.assertEqual(convert_seconds_delta_to_time(0), "0 secs")

    def test_large(self):
        self.assertEqual(convert_seconds_delta_to_time(7322), "2 hrs, 2 mins, 2 secs")
